/******************************************************************************

   This file is a part of the

   ICE-POISSON TOOLBOX: Total Variation Restoration of Images Corrupted by
   Poisson Noise with Iterated Conditional Expectations

   Copyright (C) 2018 Rémy Abergel (Remy.Abergel AT parisdescartes.fr), Cécile
   Louchet (Cecile.Louchet AT univ-orleans.fr), Lionel Moisan (Lionel.Moisan AT
   parisdescartes.fr), Tieyong Zeng (zeng AT hkbu.edu.hk).

   This software provides a C implementation of the image denoising algorithm
   described in the following research papers

   [1] R. Abergel, C. Louchet, L. Moisan and T. Zeng, ``Total Variation
   Restoration of Images Corrupted by Poisson Noise with Iterated Conditional
   Expectations''.  Proceedings of the 5th International Conference on Scale
   Space and Variational Methods in Computer Vision (SSVM), pp. 178--190, 2015,
   Springer, DOI: 10.1007/978-3-319-18461-6_15.

   [2] R. Abergel and L. Moisan, ``Fast and accurate evaluation of a generalized
   incomplete gamma function'', Preprint MAP5 2016-14,
   https://hal.archives-ouvertes.fr/hal-01329669.

   This program is free software: you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free Software
   Foundation, either version 3 of the License, or (at your option) any later
   version.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along with
   this program. If not, see <http://www.gnu.org/licenses/>.

******************************************************************************/

/* compilation (with gcc):

   + with OpenMP    : gcc -O3 ice-poisson.c tiffread.c tiffwrite.c kernel.c -lm -ltiff -fopenmp -o ice-poisson
   + without OpenMP : gcc -O3 ice-poisson.c tiffread.c tiffwrite.c kernel.c -lm -ltiff -o ice-poisson

*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

#define FREE_ALL() {					\
    if(NULL != u) free(u);				\
    if(NULL != u0) free(u0);				\
    if(NULL != u_old) free(u_old);			\
  }
#define ASSERT_ALLOC(cmd) if(NULL == (cmd)) { printf("Not enough memory\n"); FREE_ALL(); exit(EXIT_FAILURE); }

extern int read_tiff_into_double(double**,int*,int*,char*,char); /* see source file 'tiffread.c' */
extern int save_tiff(char*,double*,int,int,unsigned int16,char,char,char); /* see source file 'tiffwrite.c' */
extern void deltagammainc(double*,double*,char*,double,double,double,double); /* see source file 'kernel.c' */

/* display usage */
static void display_usage()
{
  printf("\nTV-ICE Poisson image denoising\n\n");
  printf("Usage: ice-poisson-openmp [-T numTh] [-N Niter] [-W lambda] [-f] [-v] in out\n\n");
  printf("   -T numTh  : (default = 1) number of threads (OpenMP)\n");
  printf("   -N Niter  : (default = 100) number of iterations\n");
  printf("   -W lambda : (default = 0.2) set regularity parameter (TV weight)\n");
  printf("   -f        : save output as a tiff-float instead of unsigned char tiff\n");
  printf("   -v        : light verbose mode (display progression)\n");
  printf("   in        : input image (TIFF)\n");
  printf("   out       : output image (TIFF)\n\n");
}

int main(int argc, char **argv)
{
  char *filenamein = NULL,*filenameout = NULL,*Wvalue = NULL,*Nvalue = NULL,*Tvalue = NULL,vflag,fflag,method;
  double *u0 = NULL,*u = NULL,*u_old = NULL,lambda,mu[5],v0,v1,v2,v3,v4,v5,c0,c1,c2,c3,c4,rA0,rA1,rA2,rA3,rA4,sA0,sA1,sA2,sA3,sA4,rB0,rB1,rB2,rB3,rB4,sB0,sB1,sB2,sB3,sB4,p,sAmax,sBmax,sumA,sumB,distl2;
  int niter,k,i,err,nx,ny,Omega,x,y,adr,numTh;
  vflag = fflag = 0;

  /******************************************/
  /*            arguments parser            */
  /******************************************/

  /* main parser */
  for(i=1;i<argc;i++) {
    if (strcmp(argv[i],"-v") == 0) vflag = 1;
    else if (strcmp(argv[i],"-f") == 0) fflag = 1;
    else if (strcmp(argv[i],"-T") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -T requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	Tvalue = argv[i+1]; err = sscanf(Tvalue,"%d",&numTh); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -T\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-N") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -N requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	Nvalue = argv[i+1]; err = sscanf(Nvalue,"%d",&niter); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -N\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (strcmp(argv[i],"-W") == 0) {
      if(i==argc-1) { display_usage(); printf("Error: option -W requires an argument\n\n"); return EXIT_FAILURE; }
      else {
	Wvalue = argv[i+1]; err = sscanf(Wvalue,"%lf",&lambda); i++;
	if(err != 1) { display_usage(); printf("Error: could not retrieve properly the argument of option -W\n\n"); return EXIT_FAILURE; }
      }
    }
    else if (NULL == filenamein) filenamein = argv[i];
    else if (NULL == filenameout) filenameout = argv[i];
    else { display_usage(); printf("Error: could not retrieve properly input arguments.\n\n"); return EXIT_FAILURE; }
  }

  /* set default values */
  if(NULL == Tvalue) numTh = 1;
  if(NULL == Nvalue) niter = 100;
  if(NULL == Wvalue) lambda = 0.2;

  /* check consistency */
  if ((NULL == filenamein) || (NULL == filenameout)) {
    display_usage();
    if (NULL == filenamein) printf("Error: input 'in' is missing\n\n");
    else if(NULL == filenameout) printf("Error: output 'out' is missing\n\n");
    return EXIT_FAILURE;
  }

  /* deal with verbose mode */
  if(vflag) {
    printf("\nInput parameters\n");
    printf("================\n");
    printf("number of OpenMP threads: numTh = %d\n",numTh);
    printf("number of iterations: Niter = %d\n",niter);
    printf("regularity parameter: lambda = %.17g\n\n",lambda);
  }

  /************************************************************/
  /* Open TIFF Image & get its graylevels in double precision */
  /************************************************************/
  if(vflag) {
    printf("Load input image and allocate memory\n");
    printf("====================================\n");
  }

  if(EXIT_FAILURE == read_tiff_into_double(&u0,&nx,&ny,filenamein,vflag)) {
    if(vflag) printf("Free memory and exit the program.\n\n");
    else printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /*********************/
  /* Memory allocation */
  /*********************/
  Omega = nx*ny;
  if(vflag) { printf("memory allocation: in progress"); fflush(stdout);}

  ASSERT_ALLOC(u     = (double*) malloc(Omega * sizeof(double)));
  ASSERT_ALLOC(u_old = (double*) malloc(Omega * sizeof(double)));

  if(vflag) { printf("\33[2K\rmemory allocation: done\n\n"); fflush(stdout); }


  /***********************************/
  /* KERNEL: process the input image */
  /***********************************/
  if(vflag) {
    printf("Start the image restoration process\n");
    printf("===================================\n");
  }

  /* initialize mu */
  mu[0] = 1.-4.*lambda;
  mu[1] = 1.-2.*lambda;
  mu[2] = 1.;
  mu[3] = 1.+2.*lambda;
  mu[4] = 1.+4.*lambda;

  /* set u = u_old = 0 */
#pragma omp parallel for num_threads(numTh) private(adr)
  for(adr=0;adr<Omega;adr++) u[adr] = u_old[adr] = 0.;

  /**** main loop ****/
  for(i=1;i<=niter;i++) {

    distl2 = 0.;

    // proceed each pixel //
#pragma omp parallel for num_threads(numTh) private(x,y,adr,p,v0,v1,v2,v3,v4,v5,c0,c1,c2,c3,c4,rA0,rA1,rA2,rA3,rA4,sA0,sA1,sA2,sA3,sA4,rB0,rB1,rB2,rB3,rB4,sB0,sB1,sB2,sB3,sB4,sumA,sumB,sAmax,sBmax,k) reduction (+:distl2)
    for(y=0;y<ny;y++)
      for(x=0;x<nx;x++) {

	adr = y*nx+x;

	// retrieve neighboring pixels //
	p = u0[adr];
	v1 = u_old[(x==0)?adr:adr-1];     // left neighbor pixel
	v2 = u_old[(x==nx-1)?adr:adr+1];  // right neighbor pixel
	v3 = u_old[(y==0)?adr:adr-nx];    // top neighbor pixel
	v4 = u_old[(y==ny-1)?adr:adr+nx]; // bottom neighbor pixel

	// sort neighborhood //
	while(!((v1<=v2)&&(v2<=v3)&&(v3<=v4))) {
	  if(v1>v2) { v0 = v1; v1 = v2; v2 = v0; }
	  if(v2>v3) { v0 = v2; v2 = v3; v3 = v0; }
	  if(v3>v4) { v0 = v3; v3 = v4; v4 = v0; }
	}
	v0 = 0.; v5 = INFINITY;

	// compute coefficients ci //
	c0 = lambda*( v1+v2+v3+v4);
	c1 = lambda*(-v1+v2+v3+v4);
	c2 = lambda*(-v1-v2+v3+v4);
	c3 = lambda*(-v1-v2-v3+v4);
	c4 = lambda*(-v1-v2-v3-v4);

	// numerator: compute each deltagammainc integral using a mantissa-exponent representation //
	deltagammainc(&rA0,&sA0,&method,v0,v1,mu[0],p+2.); sA0 -= c0; sAmax = sA0;
	deltagammainc(&rA1,&sA1,&method,v1,v2,mu[1],p+2.); sA1 -= c1; sAmax = fmax(sAmax,sA1);
	deltagammainc(&rA2,&sA2,&method,v2,v3,mu[2],p+2.); sA2 -= c2; sAmax = fmax(sAmax,sA2);
	deltagammainc(&rA3,&sA3,&method,v3,v4,mu[3],p+2.); sA3 -= c3; sAmax = fmax(sAmax,sA3);
	deltagammainc(&rA4,&sA4,&method,v4,v5,mu[4],p+2.); sA4 -= c4; sAmax = fmax(sAmax,sA4);

	// denominator: compute each deltagammainc integral using a mantissa-exponent representation //
	deltagammainc(&rB0,&sB0,&method,v0,v1,mu[0],p+1.); sB0 -= c0; sBmax = sB0;
	deltagammainc(&rB1,&sB1,&method,v1,v2,mu[1],p+1.); sB1 -= c1; sBmax = fmax(sBmax,sB1);
	deltagammainc(&rB2,&sB2,&method,v2,v3,mu[2],p+1.); sB2 -= c2; sBmax = fmax(sBmax,sB2);
	deltagammainc(&rB3,&sB3,&method,v3,v4,mu[3],p+1.); sB3 -= c3; sBmax = fmax(sBmax,sB3);
	deltagammainc(&rB4,&sB4,&method,v4,v5,mu[4],p+1.); sB4 -= c4; sBmax = fmax(sBmax,sB4);

	// update image at position (x,y) //
	sumA = rA0*exp(sA0-sAmax) + rA1*exp(sA1-sAmax) + rA2*exp(sA2-sAmax) + rA3*exp(sA3-sAmax) + rA4*exp(sA4-sAmax);
	sumB = rB0*exp(sB0-sBmax) + rB1*exp(sB1-sBmax) + rB2*exp(sB2-sBmax) + rB3*exp(sB3-sBmax) + rB4*exp(sB4-sBmax);
	u[adr] = exp(sAmax-sBmax) * (sumA / sumB);
	distl2 += pow(u[adr]-u_old[adr],2);
      }

    // update u_old //
#pragma omp parallel for num_threads(numTh) private(adr)
    for(adr=0;adr<Omega;adr++) u_old[adr] = u[adr];

    // display stuffs //
    if(vflag) { printf("\33[2K\riteration %d/%d, average l2 distance between u^{%d} and u^{%d} = %.4g",i,niter,i,i-1,sqrt(distl2/(double)Omega)); fflush(stdout); }
  }

  if(vflag) printf("\n\n");
  /***********************************/
  /*          END OF KERNEL          */
  /***********************************/




  /********************************/
  /* Write output as a TIFF Image */
  /********************************/
  if(vflag) {
    printf("Save restored image as a TIFF image and free memory\n");
    printf("===================================================\n");
  }

  if (EXIT_FAILURE == save_tiff(filenameout,u,nx,ny,8,fflag,1,vflag)) {
    if (vflag) printf("free memory and exit the program.\n\n");
    else printf("An error occured (try verbose mode using option -v to get more informations)\n");
    FREE_ALL();
    return EXIT_FAILURE;
  }

  /* free memory */
  if(vflag) {
    printf("free memory: in progress");
    fflush(stdout);
  }

  FREE_ALL();

  if(vflag) {
    printf("\33[2K\rfree memory: done\n\n");
    fflush(stdout);
  }

  return EXIT_SUCCESS;
}
